cimport stdlib

cdef gfxsource_t *pdfdriver
pdfdriver = gfxsource_pdf_create()

def open(doctype, filename):
    global pdfdriver
    cdef gfxsource_t *driver
    cdef gfxdocument_t *d

    if doctype == 'pdf':
        driver = pdfdriver
    else:
        raise Exception("Unknown type %s" % doctype)

    with nogil:
        d = driver.open(driver, filename)

    if not d:
        raise Exception("File does not exists")
    doc = Doc()
    doc.thisptr = d
    doc._filename = filename
    return doc

def setparameter(key, value):
    global pdfdriver
    pdfdriver.setparameter(pdfdriver, key, value)

def verbose(int val):
    """
    verbose(level)

    Set the logging verbosity of the gfx module. Log levels are:
    level=-1          Log nothing
    level=0 (fatal)   Log only fatal errors
    level=1 (error)   Log only fatal errors and errors
    level=2 (warn)    Log all errors and warnings
    level=3 (notice)  Log also some rudimentary data about the parsing/conversion
    level=4 (verbose) Log some additional parsing information
    level=5 (debug)   Log debug statements
    level=6 (trace)   Log extended debug statements

    All logging messages are written to stdout.
    """
    setConsoleLogging(val)

def addfont(filename):
    """
    addfont(filename)

    Passes an additional font file to the PDF parser. If a PDF contains
    external fonts (i.e. fonts which are not contained in the PDF itself)
    then the files added by addfont() will be searched.
    """
    global pdfdriver
    pdfdriver.setparameter(pdfdriver, "font", filename)

def addfontdir(dirname):
    """
    addfontdir(dirname)

    Passes a complete directory containing fonts to the PDF parser. Any
    font file within this directory might be used to resolve external fonts
    in PDF files
    """
    global pdfdriver
    pdfdriver.setparameter(pdfdriver, "fontdir", dirname)

cdef class Doc:
    cdef gfxdocument_t*thisptr
    cdef char*_filename

    property filename:
        def __get__(self):
            return self._filename

    property pages:
        def __get__(self):
            return self.thisptr.num_pages

    def __dealloc__(self):
        if self.thisptr:
            self.thisptr.destroy(self.thisptr)
            self.thisptr = NULL

        if self._filename:
            self._filename = NULL

    def getPage(self, int pagenr):
        page = Page()
        page.thisptr = self.thisptr.getpage(self.thisptr, pagenr)
        if not page.thisptr:
            raise Exception("Couldn't extract page %d", pagenr);
        page.parent = self

        return page

    def getInfo(self, key):
        s = self.thisptr.getinfo(self.thisptr, key)
        return s

    def setparameter(self, key, value):
        self.thisptr.setparameter(self.thisptr, key, value)

cdef class Page:
    cdef object parent
    cdef gfxpage_t*thisptr

    property size:
        def __get__(self):
            return self.thisptr.width, self.thisptr.height

    property doc:
        def __get__(self):
            return self.parent

    property nr:
        def __get__(self):
            return self.thisptr.nr

    property width:
        def __get__(self):
            return self.thisptr.width

    property height:
        def __get__(self):
            return self.thisptr.height

    def __dealloc__(self):
        if self.thisptr:
            self.thisptr.destroy(self.thisptr)
            self.thisptr = NULL

        if self.parent:
            self.parent = None

    def render(self, dev, move=None, clip=None):
        if move:
            x, y = move

        if clip:
            cx1, cy1, cx2, cy2 = clip

        cdef gfxdevice_t*output = (<OutputObject>dev).thisptr
        if x or y or cx1 or cx2 or cy1 or cy2:
            self.thisptr.rendersection(self.thisptr, output, x, y, cx1, cy1, cx2, cy2)
        else:
            self.thisptr.render(self.thisptr, output)

    def asImage(self, int width, int height, int antialise=2):
        """
        asImage(width, height, antialise=2)

        Creates a bitmap from a page. The bitmap will be returned as a string
        containing RGB triplets. The bitmap will be rescaled to the specified width and
        height. The aspect ratio of width and height doesn't need to be the same
        as the page.
        """
        if not width or not height:
            raise Exception("invalid dimensions: %dx%d" % width, height)

        cdef gfxresult_t*result
        cdef gfximage_t*img
        cdef int l, ll, s, t
        cdef char*data
        cdef gfxdevice_t dev1, dev2

        s_antialise = str(antialise)

        with nogil:
            gfxdevice_render_init(&dev1)
            dev1.setparameter(&dev1, "antialise", s_antialise)
            gfxdevice_rescale_init(&dev2, &dev1, width, height, 0)
            dev2.startpage(&dev2, <int>self.thisptr.width, <int>self.thisptr.height)

        self.thisptr.render(self.thisptr, &dev2)

        with nogil:
            dev2.endpage(&dev2);
            result = dev2.finish(&dev2)
            img = <gfximage_t*>result.get(result, "page0")
            l = img.width * img.height
            ll = l * 3
            data = <char*>stdlib.malloc(ll)
            s = 0
            t = 0
            while t < l:
                data[s+0] = img.data[t].r
                data[s+1] = img.data[t].g
                data[s+2] = img.data[t].b
                s += 3
                t += 1

            result.destroy(result)

        py_img = data[:ll]
        return py_img


cdef class OutputObject:
    cdef gfxdevice_t*thisptr

    def __cinit__(self):
        self.thisptr = <gfxdevice_t*>stdlib.malloc(sizeof(gfxdevice_t))

    def setparameter(self, key, val):
        self.thisptr.setparameter(self.thisptr, key, val)

    def startpage(self, int width, int height):
        self.thisptr.startpage(self.thisptr, width, height)

    def endpage(self):
        self.thisptr.endpage(self.thisptr)

    def save(self, filename):
        cdef gfxresult_t*result = self.thisptr.finish(self.thisptr)
        cdef int ret
        ret = result.save(result, filename)
        result.destroy(result)

        if ret < 0:
            raise Exception("Couldn't write to %s"% filename)

    def __dealloc__(self):
        stdlib.free(self.thisptr)

cdef class SWF(OutputObject):
    def __init__(self):
        gfxdevice_swf_init(self.thisptr)

cdef class OCR(OutputObject):
    def __init__(self):
        gfxdevice_ocr_init(self.thisptr)

cdef class PlainText(OutputObject):
    def __init__(self):
        gfxdevice_text_init(self.thisptr)

cdef class ImageList(OutputObject):
    def __init__(self):
        gfxdevice_render_init(self.thisptr)

cdef class PDF(OutputObject):
    def __init__(self):
        gfxdevice_pdf_init(self.thisptr)


cdef class Glyph:
    cdef gfxglyph_t *thisptr

    property advance:
        def __get__(self):
            return self.thisptr.advance

    property unicode:
        def __get__(self):
            return self.thisptr.unicode

    cdef __setptr(self, gfxglyph_t *glyph):
        self.thisptr = <gfxglyph_t*>glyph

cdef class Font:
    cdef gfxfont_t *thisptr
    cdef object py_unicode2glyph
    cdef object py_glyphs

    property id:
        def __get__(self):
            return self.thisptr.id

    property num_glyphs:
        def __get__(self):
            return self.thisptr.num_glyphs

    property max_unicode:
        def __get__(self):
            return self.thisptr.max_unicode

    property ascent:
        def __get__(self):
            return self.thisptr.ascent

    property descent:
        def __get__(self):
            return self.thisptr.descent

    property unicode2glyph:
        def __get__(self):
            return self.py_unicode2glyph

    property kerning:
        def __get__(self):
            return self.thisptr.kerning.c1, self.thisptr.kerning.c2, self.thisptr.kerning.advance

    property kerning_size:
        def __get__(self):
            return self.thisptr.kerning_size

    property glyphs:
        def __get__(self):
            return self.py_glyphs

    cdef __setptr(self, gfxfont_t *font):
        cdef int i
        self.thisptr = <gfxfont_t*>font
        self.py_unicode2glyph = [self.thisptr.unicode2glyph[i] for i in range(0, self.thisptr.max_unicode)]
        self.py_glyphs = []
        for i in range(0, self.thisptr.num_glyphs):
            gly = Glyph()
            gly.__setptr(&self.thisptr.glyphs[i])
            self.py_glyphs.append(gly)
       


cdef object __font(gfxfont_t*font):
    f = Font()
    f.__setptr(font)
    return f

cdef int _my_setparameter(gfxdevice_t*dev, const_char_ptr key, const_char_ptr value) with gil:
    cdef PassThrough self = <object><void*>dev.internal
    f = getattr(self.obj, 'setparameter', None)
    if f:
        f(key, value)
    return 1
cdef int my_setparameter(gfxdevice_t*dev, const_char_ptr key, const_char_ptr value) nogil:
    return _my_setparameter(dev, key, value)

cdef void _my_startpage(gfxdevice_t*dev, int width, int height) with gil:
    cdef PassThrough self = <object><void*>dev.internal
    f = getattr(self.obj, 'startpage', None)
    if f:
        f(width, height)
cdef void my_startpage(gfxdevice_t*dev, int width, int height) nogil:
    _my_startpage(dev, width, height)


MOVETO = gfx_moveTo
LINETO = gfx_lineTo
SPLINETO = gfx_splineTo
cdef object __line_type(gfxline_t*line):
        return line.type, (line.x, line.y, line.sx, line.sy)

cdef class Gradient:
    cdef gfxgradient_t*thisptr

    property color:
        def __get__(self):
            return __color(<gfxcolor_t*>&self.thisptr.color)

    property pos:
        def __get__(self):
            return self.thisptr.pos

    def __cinit__(self, gradient):
        self.thisptr = <gfxgradient_t*>gradient
        self.pos = self.thisptr.pos

    def __iter__(self):
        return self

    def next(self):
        if self.thisptr.next is NULL:
            raise StopIteration
        self.thisptr = self.thisptr.next

cdef object __gradient(gfxgradient_t*gradient):
        klass = Gradient()
        return klass

cdef object __color(gfxcolor_t*color):
    return color.r, color.g, color.b, color.a

cdef void my_startclip(gfxdevice_t*dev, gfxline_t*line):
    cdef PassThrough self = <object><void*>dev.internal
    f = getattr(self.obj, 'startclip', None)
    if f:
        klass = __line_type(line)
        f(klass)

cdef void my_addfont(gfxdevice_t*dev, gfxfont_t*font):
    cdef PassThrough self = <object><void*>dev.internal
    f = getattr(self.obj, 'addfont', None)
    if f:
        kfont = __font(font)
        f(kfont)

cdef void my_endclip(gfxdevice_t*dev):
    cdef PassThrough self = <object><void*>dev.internal
    f = getattr(self.obj, 'endclip', None)
    if f:
        f()

CAP_BUTT = gfx_capButt
CAP_ROUND = gfx_capRound
CAP_SQUARE = gfx_capSquare
JOIN_MITER = gfx_joinMiter
JOIN_ROUND = gfx_joinRound
JOIN_BEVEL = gfx_joinBevel
cdef void my_stroke(gfxdevice_t*dev, gfxline_t*line, gfxcoord_t width, gfxcolor_t*color, gfx_capType cap_style, gfx_joinType joint_style, gfxcoord_t miterLimit):
    cdef PassThrough self = <object><void*>dev.internal
    f = getattr(self.obj, 'stroke', None)
    if f:
        klass = __line_type(line)
        kcolor = __color(color)
        f(klass, width, kcolor, cap_style, joint_style, miterLimit)

cdef void my_fill(gfxdevice_t*dev, gfxline_t*line, gfxcolor_t*color):
    cdef PassThrough self = <object><void*>dev.internal
    f = getattr(self.obj, 'fill', None)
    if f:
        klass = __line_type(line)
        kcolor = __color(color)
        f(klass, kcolor)

cdef object __matrix(gfxmatrix_t*matrix):
    return ((matrix.m00, matrix.m10, matrix.tx), (matrix.m01, matrix.m11, matrix.ty))

cdef class ImageObject:
    cdef gfximage_t *thisptr
    cdef char*_data

    property width:
        def __get__(self):
            return self.thisptr.width
    property height:
        def __get__(self):
            return self.thisptr.height
    property data:
        def __get__(self):
            cdef int l, s, t
            cdef long ll
            l = self.thisptr.width * self.thisptr.height
            ll = l * 3
            if self._data is NULL:
                s = t = 0
                self._data = <char*>stdlib.malloc(ll)
                while t < l:
                    self._data[s+0] = self.thisptr.data[t].r
                    self._data[s+1] = self.thisptr.data[t].g
                    self._data[s+2] = self.thisptr.data[t].b
                    s += 3
                    t += 1
            py_img = self._data[:ll]
            return py_img

    def __cinit__(self, image):
        self.thisptr = <gfximage_t*>image
        self._data = NULL

    def __dealloc__(self):
        if self._data:
            stdlib.free(self._data)

cdef object __cxform(gfxcxform_t*cxform):
    if cxform is NULL:
        return None
    return ((cxform.rr, cxform.rg, cxform.ra, cxform.tr),
            (cxform.gr, cxform.gg, cxform.ga, cxform.tg),
            (cxform.br, cxform.bg, cxform.ba, cxform.tb),
            (cxform.ar, cxform.ag, cxform.aa, cxform.ta),
           )

cdef void my_fillbitmap(gfxdevice_t*dev, gfxline_t*line, gfximage_t*img, gfxmatrix_t*imgcoord2devcoord, gfxcxform_t*cxform):
    cdef PassThrough self = <object><void*>dev.internal
    f = getattr(self.obj, 'fillbitmap', None)
    if f:
        klass = __line_type(line)
        kmatrix = __matrix(imgcoord2devcoord)
        kimg = ImageObject(<object>img)
        kcxform = __cxform(cxform)
        f(klass, kimg, kmatrix, kcxform)

GRADIENT_RADIAL = gfxgradient_radial
GRADIENT_LINEAR = gfxgradient_linear
cdef void my_fillgradient(gfxdevice_t*dev, gfxline_t*line, gfxgradient_t*gradient, gfxgradienttype_t gtype, gfxmatrix_t*matrix):
    cdef PassThrough self = <object><void*>dev.internal
    f = getattr(self.obj, 'fillgradient', None)
    if f:
        klass = __line_type(line)
        kgrad = __gradient(gradient)
        kmatrix = __matrix(matrix)
        f(klass, kgrad, gtype, kmatrix)

cdef void my_drawchar(gfxdevice_t*dev, gfxfont_t*font, int glyphnr, gfxcolor_t*color, gfxmatrix_t*matrix):
    cdef PassThrough self = <object><void*>dev.internal
    f = getattr(self.obj, 'drawchar', None)
    if f:
        kcolor = __color(color)
        kfont = __font(font)
        kmatrix = __matrix(matrix)
        f(kfont, glyphnr, kcolor, kmatrix)

cdef void my_drawlink(gfxdevice_t*dev, gfxline_t*line, const_char_ptr action):
    cdef PassThrough self = <object><void*>dev.internal
    f = getattr(self.obj, 'drawlink', None)
    if f:
        klass = __line_type(line)
        f(klass, action)

cdef void _my_endpage(gfxdevice_t*dev) with gil:
    cdef PassThrough self = <object><void*>dev.internal
    f = getattr(self.obj, 'endpage', None)
    if f:
        f()
cdef void my_endpage(gfxdevice_t*dev) nogil:
    _my_endpage(dev)

cdef _my_finish(gfxdevice_t*dev) with gil:
    cdef PassThrough self = <object><void*>dev.internal
    f = getattr(self.obj, 'finish', None)
    if f:
        f()

cdef gfxresult_t* my_finish(gfxdevice_t*dev) nogil:
     _my_finish(dev)

cdef class PassThrough(OutputObject):
    cdef object obj
    def __cinit__(self, obj):
        self.obj = obj
        self.thisptr.name = "passthrough"

        self.thisptr.setparameter = my_setparameter
        self.thisptr.startpage = my_startpage
        self.thisptr.startclip = my_startclip
        self.thisptr.addfont = my_addfont
        self.thisptr.endclip = my_endclip
        self.thisptr.stroke = my_stroke
        self.thisptr.fill = my_fill
        self.thisptr.fillbitmap = my_fillbitmap
        self.thisptr.fillgradient = my_fillgradient
        self.thisptr.drawchar = my_drawchar
        self.thisptr.drawlink = my_drawlink
        self.thisptr.endpage = my_endpage
        self.thisptr.finish = my_finish
        self.thisptr.internal = <void*>self

    def __init__(self, obj):
        pass
